package com.tomtomwork.queueing.test.jmsintegrationtest;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import javax.jms.Connection;
import javax.jms.JMSException;
import javax.jms.Session;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;

/**
 * JMS connector
 */

abstract class JMSConnector extends GenericConnector
{
 // === Static ================================================================

 private static final Logger LOGGER = Logger.getLogger(JMSConnector.class); 
 
 // === Attributes ============================================================
 
          Connection          connection;
          Session             session;

 final    String              jndiFactoryInitial, jndiProviderURL, jndiConnectionFactoryAlias, jndiQueueAlias;
 final    Map<String, String> addJndiICProperties;

 // === Constructor ===========================================================
       
 /**
  * Create a new connector instance
  * @param name                       the connector name
  * @param jndiFactoryInitial         the JNDI factory alias
  * @param jndiProviderURL            the JNDI provider URL
  * @param jndiConnectionFactoryAlias the JNDI connection factory alias
  * @param jndiQueueAlias             the JNDI queue alias
  * @param addJndiICProperties        any additional JNDI InitialContext properties
  */
       
 public JMSConnector(final String name,
                     final String jndiFactoryInitial, final String jndiProviderURL, final String jndiConnectionFactoryAlias, final String jndiQueueAlias,
                     final Map<String, String> addJndiICProperties)
 {
  super(name);
  
  // Set properties
  
  this.jndiFactoryInitial         = jndiFactoryInitial;
  this.jndiProviderURL            = jndiProviderURL;
  this.jndiConnectionFactoryAlias = jndiConnectionFactoryAlias;
  this.jndiQueueAlias             = jndiQueueAlias;
  this.addJndiICProperties        = Collections.unmodifiableMap(addJndiICProperties);

  // Log parameters

  LOGGER.info("=== " + this.getClass().getSimpleName() + " :: " + name + " ===");
  LOGGER.info("JNDIFACTORYINITIAL         = " + jndiFactoryInitial);
  LOGGER.info("JNDIPROVIDERURL            = " + jndiProviderURL);
  LOGGER.info("JNDICONNECTIONFACTORYALIAS = " + jndiConnectionFactoryAlias);
  LOGGER.info("JNDIQUEUEALIAS             = " + jndiQueueAlias);
  LOGGER.info("ADDJNDIICPROPERTIES        = " + addJndiICProperties);
 }
 
 /**
  * Create a new connector instance
  * @param name                 the connector name
  * @param connectionProperties a file reference pointing to a properties file containing
  *                             the connector properties
  */
 
 public JMSConnector(final String name, final File connectionProperties)
 throws IOException
 {
  super(name);  
  
  Properties          properties;
  Map<String, String> addJndiICProperties;

  // Load properties

  properties          = Util.loadProperties(connectionProperties);
  addJndiICProperties = new HashMap<String, String>();

  // Set properties

  this.jndiFactoryInitial         = properties.getProperty("jndi.ic.factoryinitial",       null);
  this.jndiProviderURL            = properties.getProperty("jndi.ic.providerurl",          null);
  this.jndiConnectionFactoryAlias = properties.getProperty("jndi.alias.connectionfactory", null);
  this.jndiQueueAlias             = properties.getProperty("jndi.alias.queue",             null);

  for(String key : properties.stringPropertyNames())
  {
   if(key.startsWith("jndi.ic.add.")) addJndiICProperties.put(key.substring("jndi.ic.add.".length()), properties.getProperty(key));
  }

  this.addJndiICProperties = Collections.unmodifiableMap(addJndiICProperties);

  // Log parameters

  LOGGER.info("=== " + this.getClass().getSimpleName() + " :: " + name + " ===");
  LOGGER.info("JNDIFACTORYINITIAL         = " + jndiFactoryInitial);
  LOGGER.info("JNDIPROVIDERURL            = " + jndiProviderURL);
  LOGGER.info("JNDICONNECTIONFACTORYALIAS = " + jndiConnectionFactoryAlias);
  LOGGER.info("JNDIQUEUEALIAS             = " + jndiQueueAlias);
  LOGGER.info("ADDJNDIICPROPERTIES        = " + addJndiICProperties);
 }

 // === Public ================================================================
 
 /**
  * @see com.tomtomwork.queueing.test.jmsintegrationtest.GenericConnector#shutdown()
  */
 
 void shutdown()
 {
  disconnect();
 }
 
 /**
  * Connect the connector
  * @throws JMSException
  * @throws NamingException
  */
 
 abstract void connect()
 throws JMSException,
        NamingException;

 /**
  * Connect the connector with retries
  * @throws InterruptedException
  */
 
 void connectWithRetries()
 throws InterruptedException
 {
  // Loop until successful connect
  
  for(;;)
  {
   try
   {
    connect(); return;
   }
   catch(Exception e)
   {
    LOGGER.error(name + ": Exception while connecting; retrying.", e);

    disconnect();
    Thread.sleep(1000);
   }
  }
 }

 /**
  * Disconnect the connector
  */
 
 abstract void disconnect();

 /**
  * Reconnect the connector with retries
  * @throws InterruptedException
  */
 
 void reconnectWithRetries()
 throws InterruptedException
 {
  disconnect();
  connectWithRetries();
 }
 
 /**
  * @return
  * @throws NamingException
  */
 
 InitialContext createInitialContext() 
 throws NamingException
 {
  Properties  jndiProperties;

  // Initialize JNDI

  jndiProperties = new Properties();

  // Add standard properties

  jndiProperties.setProperty(Context.INITIAL_CONTEXT_FACTORY, jndiFactoryInitial);
  jndiProperties.setProperty(Context.PROVIDER_URL,            jndiProviderURL);

  // Add additional properties

  for(Map.Entry<String, String> entry : addJndiICProperties.entrySet())
  {
   jndiProperties.setProperty(entry.getKey(), entry.getValue());
  }

  // Create and return context
  
  return(new InitialContext(jndiProperties));
 }  
 
 /**
  * Stop the connector
  */
 
 void signalStop()
 {
  stop = true;
 }  
   
 // === Package protected =====================================================
 // === Protected =============================================================
 // === Private ===============================================================
 // === Inner classes =========================================================   
}