package com.tomtomwork.queueing.test.jmsintegrationtest;

import java.io.File;
import java.io.IOException;

import java.util.Map;

import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MessageConsumer;
import javax.jms.TextMessage;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;

/**
 * Message bridge
 * @author rschuetz
 */

public class MessageBridge
{
 // === Static ================================================================
 
 private static final Logger         LOGGER          = Logger.getLogger(MessageGenerator.class);
 
 private static final long           RECEIVE_TIMEOUT  = 1L * 60L * 1000,
                                     MSG_LOG_INTERVAL = 10L * 1000L;
 
 // === Attributes ============================================================ 
 
 private final File                  STOPFILE;
 
 private       Bridge                bridge;
 
 // === Constructor ===========================================================

 /**
  * Create a new MessageBridge instance
  * @param args the command line arguments
  * @throws IOException 
  */
 
 public MessageBridge(String[] args) 
 throws IOException
 {
  int    argIdx;
  String bridgePropertiesFile;
 
  // Get parameters

  if(args.length == 2)
  {
   argIdx               = 0;

   bridgePropertiesFile =                                                                             args[argIdx ++];
   STOPFILE             = new File(                                                                   args[argIdx ++]);
  }
  else
  {
   throw new IllegalArgumentException("Usage: " + MessageBridge.class.getSimpleName() +
                                      " <bridge properties file> <stop file name>");
  }

  // Log parameters

  LOGGER.info("=== " + this.getClass().getSimpleName() + " ===");
  LOGGER.info("BRIDGECONFIG  = " + bridgePropertiesFile);
  LOGGER.info("STOPFILE      = " + STOPFILE);

  // Create bridge

  bridge = new Bridge("B", new File(bridgePropertiesFile));
 }
 
 // === Public ================================================================ 
 
 /**
  * @param args the command line arguments
  * @throws Exception
  */
 
 public static void main(String[] args)
 throws Exception
 {
  (new MessageBridge(args)).runLoadTest();
  System.exit(0);
 } 
 
 /**
  * Run the load test
  */
 
 public void runLoadTest()
 {
  try
  {
   // Launch threads

   bridge.start();

   // Loop 
   
   while(true)
   {
    // Check for stop file

    if(STOPFILE.exists())
    {
     LOGGER.info("Stop file found; stopping ...");
     break;
    }
    
    // Sleep
    
    Thread.sleep(100);
   }
   
   // Signal threads to stop
   
   bridge.signalStop();
   
   // Join
   
   bridge.join();
  }
  catch(Exception e)
  {
   LOGGER.error("Exception caught", e);
  }
  finally
  {
   try
   {
    bridge.shutdown();
   }
   catch(Exception e)
   {
    LOGGER.error("Exception caught while disconnecting", e);
   }
  }
 }
 
 // === Package protected =====================================================
 // === Protected =============================================================
 // === Private ===============================================================

 /**
  * Bridge class
  */
 
 private class Bridge extends JMSConnector 
 {
  // === Static ================================================================
  // === Attributes ============================================================

  private MessageConsumer consumer;
  
  // === Constructor ===========================================================
  
  /**
   * Create a new Receiver instance
   * @param name                       the connector name
   * @param jndiFactoryInitial         the JNDI factory alias
   * @param jndiProviderURL            the JNDI provider URL
   * @param jndiConnectionFactoryAlias the JNDI connection factory alias
   * @param jndiQueueAlias             the JNDI queue alias
   * @param addJndiICProperties        any additional JNDI InitialContext properties
   */
  
  public Bridge(final String name, 
                final String jndiFactoryInitial, final String jndiProviderURL, final String jndiConnectionFactoryAlias, final String jndiQueueAlias,
                final Map<String, String> addJndiICProperties)
  {
   // Call super constructor
   
   super(name, jndiFactoryInitial, jndiProviderURL, jndiConnectionFactoryAlias, jndiQueueAlias, addJndiICProperties);
  }
  
  /**
   * Create a new Bridge
   * @throws IOException 
   * @see JMSConnector#Connector(String, File)
   */
  
  public Bridge(final String name, final File connectionProperties) 
  throws IOException
  {
   // Call super constructor   
   
   super(name, connectionProperties);
  }

  // === Public ================================================================
  
  /**
   * Send messages until stopped.
   * @see java.lang.Runnable#run()
   */
  
  public void run()
  {
   boolean      retryBatch;
   int          retryID;
   long         receivedMsgSequenceID, lastMsgLogTime;
   String       messageText;
   
   TextMessage  textMessage;   
   
   // Initialize
   
   lastMsgLogTime = 0L;

   // Try/Catch
   
   try
   {
    // Initial connect
   
    connectWithRetries();
    
    // Try/Catch for InterruptedExceptions
    
    try
    {
     // Loop until stopped
     
     while(!stop)
     {
      // Loop until a batch has been received and committed. Batch is being retried
      // in case of message creation, sending or commit failures. Duplicates can
      // be generated when retrying after commit failures.
      
      for(retryID = 0; ; retryID ++)
      {
       // Initialize
       
       retryBatch = false;
       
       // Receive message
      
       try
       {
        // Receive
        
        if((textMessage = ((TextMessage)(consumer.receive(RECEIVE_TIMEOUT)))) != null)
        {
         // Get message text
         
         messageText = textMessage.getText();
         
         // Get message sequence ID of received message
         
         receivedMsgSequenceID = Util.getMessageID(messageText);

         // Log?
         
         if(lastMsgLogTime < System.currentTimeMillis() - MSG_LOG_INTERVAL)
         {
          LOGGER.info(name + ": Message # " + numberFormatter.format(receivedMsgSequenceID) + " received.");
          lastMsgLogTime = System.currentTimeMillis();
         }
        }
        else
        {
         // No message received within timeout. Throw exception to reconnect and retry.
         
         throw new JMSException(name + ": Timed out receiving messages.");
        }
       }
       catch(JMSException e)
       {
        // Message reception failed.
        // Log
        
        LOGGER.error(name + ": Message reception failed, reconnecting and retrying.", e);
         
        // Retry batch
         
        retryBatch = true;
       }
       
       // Commit batch
       
       if(!retryBatch)
       {
        // Try to commit
        
        try
        {
         // Commit
         
         session.commit();
        }
        catch(JMSException e)
        {
         // Log
         
         LOGGER.error(name + ": Commit failed, reconnecting and retrying.", e);
          
         // Retry batch
          
         retryBatch = true;
        }        
       }
       
       // Retry? Reconnect first. Else break loop
       
       if(retryBatch) reconnectWithRetries();      
                 else break;
      }
      
      // Pause?
      
      while(pause && !stop) Thread.sleep(100);
     }
    }
    catch(InterruptedException e)
    {
     // Log
       
     LOGGER.error(name + ": Caught exception", e);
    }
   }
   catch(InterruptedException e)
   {
    // Log
    
    LOGGER.error(name + ": Interrupted", e);    
   }    
   finally
   {
    disconnect();
   }
  }

  /**
   * @see com.tomtomwork.webfleet.components.JMSConnector.test.MultiThreadedLoadTest.Connector#connect()
   */
  
  @Override
  public void connect()
  throws JMSException,
         NamingException
  {
   InitialContext    jndiContext;
   ConnectionFactory connectionFactory;
   Destination       destination;

   // Log

   LOGGER.info(name + ": Connecting.");

   // Initialize context
   
   jndiContext = createInitialContext();

   // Lookup connection factory and queue

   connectionFactory = ((ConnectionFactory)(jndiContext.lookup(jndiConnectionFactoryAlias)));
   destination       = ((Destination      )(jndiContext.lookup(jndiQueueAlias)));

   // Create producer connection, session and message consumer

   connection = connectionFactory.createConnection();
   session    = connection.createSession(true, 0);

   consumer = session.createConsumer(destination);

   connection.start();

   // Log

   LOGGER.info(name + ": Connected.");
  }

  /**
   * @see com.tomtomwork.webfleet.components.JMSConnector.test.MultiThreadedLoadTest.Connector#disconnect()
   */
  
  @Override
  public void disconnect()
  {
   // Log

   LOGGER.info(name + ": Disconnecting.");

   // Close assets

   if(consumer   != null) { try { consumer  .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting consumer",   e); } consumer   = null; }
   if(session    != null) { try { session   .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting session",    e); } session    = null; }
   if(connection != null) { try { connection.close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting connection", e); } connection = null; }

   // Log

   LOGGER.info(name + ": Disconnected.");
  }
  
  // === Package protected =====================================================
  // === Protected =============================================================
  // === Private ===============================================================
  // === Inner classes =========================================================   
 }
}