package com.tomtomwork.queueing.test.jmsintegrationtest;

import java.io.File;
import java.io.IOException;


import java.util.Map;

import javax.jms.ConnectionFactory;
import javax.jms.DeliveryMode;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MessageProducer;
import javax.jms.TextMessage;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;

/**
 * Message generator
 * @author rschuetz
 */

public class MessageGenerator
{
 // === Static ================================================================
 
 private static final Logger LOGGER           = Logger.getLogger(MessageGenerator.class);
 
 private static final long   MSG_LOG_INTERVAL = 10L * 1000L;
 
 // === Attributes ============================================================ 
 
 private final File   STOPFILE;
 private final int    DELAY;

 private       Sender sender;
 
 // === Constructor ===========================================================

 /**
  * Create a new MessageGenerator instance
  * @param args the command line arguments
  */
 
 public MessageGenerator(String[] args)
 throws IOException
 {
  int    argIdx;
  String senderPropertiesFile;

  // Initialize logger

  BasicConfigurator.resetConfiguration();

  Logger.getRootLogger().addAppender(new ConsoleAppender(new PatternLayout("%-5p [%d{ISO8601}] %m%n")));
  Logger.getRootLogger().setLevel(Level.INFO);

  // Get parameters

  if(args.length == 3)
  {
   argIdx                = 0;

   senderPropertiesFile  =                  args[argIdx ++];

   DELAY                 = Integer.parseInt(args[argIdx ++]);
   STOPFILE              = new File        (args[argIdx ++]);
  }
  else
  {
   throw new IllegalArgumentException("Usage: " + MessageGenerator.class.getSimpleName() +
                                      " <sender properties file> <delay between msgs in ms> <stop file name>");
  }

  // Log parameters

  LOGGER.info("=== " + this.getClass().getSimpleName() + " ===");
  LOGGER.info("SENDERCONFIG = " + senderPropertiesFile);
  LOGGER.info("DELAY        = " + DELAY);
  LOGGER.info("STOPFILE     = " + STOPFILE);

  // Create sender

  sender = new Sender("S", new File(senderPropertiesFile));
 }
 
 // === Public ================================================================ 
 
 /**
  * @param args the command line arguments
  * @throws Exception
  */
 
 public static void main(String[] args)
 throws Exception
 {
  (new MessageGenerator(args)).runLoadTest();
  System.exit(0);
 } 
 
 /**
  * Run the load test
  */
 
 public void runLoadTest()
 {
  try
  {
   // Launch threads

   sender.start();

   // Loop 
   
   while(true)
   {
    // Check for stop file

    if(STOPFILE.exists())
    {
     LOGGER.info("Stop file found; stopping ...");
     break;
    }
    
    // Sleep
    
    Thread.sleep(100);
   }
   
   // Signal threads to stop
   
   sender.signalStop();
   
   // Join
   
   sender.join();
  }
  catch(Exception e)
  {
   LOGGER.error("Exception caught", e);
  }
  finally
  {
   try
   {
    sender.shutdown();
   }
   catch(Exception e)
   {
    LOGGER.error("Exception caught while disconnecting", e);
   }
  }
 }
 
 // === Package protected =====================================================
 // === Protected =============================================================
 // === Private ===============================================================
 // === Inner classes =========================================================

 /**
  * Sender class
  */
 
 private class Sender extends JMSConnector
 {
  // === Static ================================================================
  // === Attributes ============================================================
  
  private MessageProducer producer;

  // === Constructor ===========================================================
  
  /**
   * Create a new Sender instance
   * @param name                       the connector name
   * @param jndiFactoryInitial         the JNDI factory alias
   * @param jndiProviderURL            the JNDI provider URL
   * @param jndiConnectionFactoryAlias the JNDI connection factory alias
   * @param jndiQueueAlias             the JNDI queue alias
   * @param addJndiICProperties        any additional JNDI InitialContext properties
   */
  
  public Sender(final String name,
                final String jndiFactoryInitial, final String jndiProviderURL, final String jndiConnectionFactoryAlias, final String jndiQueueAlias,
                final Map<String, String> addJndiICProperties)
  {
   super(name, jndiFactoryInitial, jndiProviderURL, jndiConnectionFactoryAlias, jndiQueueAlias, addJndiICProperties);
  }
  
  /**
   * Create a new Sender instance
   * @see JMSConnector#Connector(String, File)
   */
  
  public Sender(final String name, final File connectionProperties)
  throws IOException
  {
   super(name, connectionProperties);
  }

  // === Public ================================================================
  
  /**
   * Send messages until stopped.
   * @see java.lang.Runnable#run()
   */
  
  public void run()
  {
   boolean     retryBatch;
   int         retryID;
   long        msgSequenceID, preBatchMsgSequenceID, lastMsgLogTime;
   
   TextMessage textMessage;   

   // Initialize
   
   lastMsgLogTime = 0L;
   msgSequenceID  = 0L;
   
   try
   {
    // Initial connect
   
    connectWithRetries();
   
    // Try/Catch for InterruptedExceptions
    
    try
    {
     // Loop until stopped
     
     while(!stop)
     {
      // Save message sequence ID
      
      preBatchMsgSequenceID = msgSequenceID;     
      
      // Loop until a batch has been sent and committed. Batch is being retried
      // in case of message creation, sending or commit failures. Duplicates can
      // be generated when retrying after commit failures.
      
      for(retryID = 0; ; retryID ++)
      {
       // Initialize
       
       retryBatch = false;
       
       // Set initial message sequence ID
       
       msgSequenceID = preBatchMsgSequenceID;
       
       // Create & send message
      
       try
       {
        // Create and send message
         
        textMessage = session.createTextMessage(Util.createRandomMessage(msgSequenceID));
        producer.send(textMessage);        
         
        // Log?
         
        if(lastMsgLogTime < System.currentTimeMillis() - MSG_LOG_INTERVAL)
        {
         LOGGER.info(name + ": Message # " + numberFormatter.format(msgSequenceID) + " sent.");
         lastMsgLogTime = System.currentTimeMillis();
        }
         
        // Increase message sequence ID
         
        msgSequenceID ++;
         
        // Impose delay?
         
        if(DELAY > 0) Thread.sleep(DELAY);
       }
       catch(JMSException e)
       {
        // Message creation or sending failed.
        // Log
        
        LOGGER.error(name + ": Message creation or sending failed, reconnecting and retrying.", e);
         
        // Retry batch
         
        retryBatch = true;
       }
       
       // Commit batch
       
       if(!retryBatch)
       {
        // Try to commit
        
        try
        {
         // Commit
         
         session.commit();
        }
        catch(JMSException e)
        {
         // Log
         
         LOGGER.error(name + ": Commit failed, reconnecting and retrying.", e);
         
         // Retry batch
          
         retryBatch = true;
        }        
       }
       
       // Retry? Reconnect first. Else break loop
       
       if(retryBatch) reconnectWithRetries();      
                 else break;
      }
      
      // Pause?
      
      while(pause && !stop) Thread.sleep(100);
     }
    }
    catch(InterruptedException e)
    {
     // Log
       
     LOGGER.error(name + ": Caught exception", e);
    }
   }
   catch(InterruptedException e)
   {
    // Log
    
    LOGGER.error(name + ": Interrupted", e);    
   }
   finally
   {
    disconnect();
   }
  }

  /**
   * @see com.tomtomwork.webfleet.components.JMSConnector.test.MultiThreadedLoadTest.Connector#connect()
   */
  
  @Override
  public void connect()
  throws JMSException,
         NamingException
  {
   InitialContext    jndiContext;
   ConnectionFactory connectionFactory;
   Destination       destination;

   // Log

   LOGGER.info(name + ": Connecting.");

   // Initialize context
   
   jndiContext = createInitialContext();

   // Lookup connection factory and queue

   connectionFactory = ((ConnectionFactory)(jndiContext.lookup(jndiConnectionFactoryAlias)));
   destination       = ((Destination      )(jndiContext.lookup(jndiQueueAlias)));

   // Create producer connection, session and message producer

   connection = connectionFactory.createConnection();
   session    = connection.createSession(true, 0);

   producer = session.createProducer(destination);
   producer.setDeliveryMode(DeliveryMode.PERSISTENT);

   connection.start();

   // Log

   LOGGER.info(name + ": Connected.");
  }

  /**
   * @see com.tomtomwork.webfleet.components.JMSConnector.test.MultiThreadedLoadTest.Connector#disconnect()
   */
  
  @Override
  public void disconnect()
  {
   // Log

   LOGGER.info(name + ": Disconnecting.");

   // Close assets

   if(producer   != null) { try { producer  .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting producer",   e); } producer   = null; }
   if(session    != null) { try { session   .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting session",    e); } session    = null; }
   if(connection != null) { try { connection.close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting connection", e); } connection = null; }

   // Log

   LOGGER.info(name + ": Disconnected.");
  }
  
  // === Package protected =====================================================
  // === Protected =============================================================
  // === Private ===============================================================
  // === Inner classes =========================================================   
 }
}